(ns puppetlabs.rbac.sync
  (:require [clojure.java.jdbc :as jdbc]
            [clojure.set :refer [difference]]
            [puppetlabs.rbac.storage :as st]
            [puppetlabs.rbac.ldap :as ldap]))

(defn update-remote-user-groups! [db remote-user]
  "Updates remote user group_ids. Takes a database spec and
  map matching the schema for a remote user."
  (let [id   (:id remote-user)
        gids (:group_ids remote-user)]
    (st/update-associated-ids! db :groupings :ruid id :gid gids)))

(defn maybe-update-groups!
  "Takes a database spec, a list of groups we had recorded in the database,
  and a list of the groups we learned about while checking a user's membership
  over LDAP. Updates the display names of any of the groups if the list in new
  groups has changed."
  [db old-groups new-groups]
  (doseq [new-group new-groups]
    (when-let [old-group (first (filter #(= (:login new-group) (:login %)) old-groups))]
      (when-not (= (:display_name old-group) (:display_name new-group))
        (jdbc/update! db :subjects {:display_name (:display_name new-group)} ["login = ?" (:login old-group)])))))

(defn maybe-delete-groups!
  ;; TODO This function is left behind; we decided against deleting groups. See PE-4994.
  "Takes a database spec and a list of group logins. For any groups that a user no longer belongs to,
  checks DS to make sure that group still exists. If not, deletes the group from the database."
  [db group-logins]
  (let [settings (st/get-directory-settings db)]
    (doseq [login group-logins]
        (when-not (ldap/group-info settings login)
          (jdbc/delete! db :subjects ["login = ?" login])))))

(defn maybe-update-remote-user!
  "Takes a database spec and a map matching the schema for a remote user.
  Updates the display name and email of the remote user if they have changed."
  [db remote-user]
  (let [settings     (st/get-directory-settings db)
        ldap-subject (ldap/user-info settings (:login remote-user))]
    (when-not (and (= (:display_name ldap-subject) (:display_name remote-user))
                   (= (:email ldap-subject) (:email remote-user)))
      (jdbc/update! db :subjects (select-keys ldap-subject [:display_name :email]) ["id = ?" (:id remote-user)]))))

(defn maybe-update-groups-for-user!
  "Takes a database spec and a map matching the schema for a remote user.
  Finds a user's current groups in the DS (filtered to only include groups
  already in the database), updates the user's group associations, and updates
  information on those groups belonging to the user in the DS."
  [db remote-user]
  (let [settings     (st/get-directory-settings db)
        known-groups (st/get-groups db)]
    (when-not (empty? known-groups)
      (let [known-group-logins (map :login known-groups)
            ldap-groups        (ldap/groups-for-user settings known-group-logins (:login remote-user))
            ;; Note, we could also filter known-groups to get these
            ;; IDs. It skips the database but does more
            ;; computation. In the future this might be a place for an
            ;; optimization.
            new-groups         (map #(st/get-subject db :login (:login %)) ldap-groups)
            new-group-ids      (map :id new-groups)]
        (update-remote-user-groups! db (assoc remote-user :group_ids new-group-ids))
        (maybe-update-groups! db new-groups ldap-groups)))))

(defn sync-remote-user!
  "Takes a database spec and the id of a remote user. Syncs the remote user's
  data and group membership information with the directory service."
  [db id]
  (let [subject-map (st/get-subject db :id id)]
    (maybe-update-groups-for-user! db subject-map)
    (maybe-update-remote-user! db subject-map)))

(defn create-remote-user!
  "Takes a database spec and the login of the remote user to be created.
  Fetches information on the user from the directory service, then adds
  the user to the database, along with its group membership information."
  [db login]
  (let [settings (st/get-directory-settings db)]
    (if-let [data (ldap/user-info settings login)]
      (let [user-data   (select-keys data [:login :display_name :email])
            _           (st/create-remote-user! db user-data)
            subject-map (st/get-subject db :login login)]
        (maybe-update-groups-for-user! db subject-map))
      (throw (IllegalArgumentException. (format "No such user found via LDAP: %s" login))))))

(defn create-group!
  "Tales a database spec and a map that matches the PostGroup schema.
  Fetches information on the group from the directory service, then adds
  the group to the database."
  [db new-group]
  (let [settings (st/get-directory-settings db)]
    (if-let [ds-data (ldap/group-info settings (:login new-group))]
      (let [group-data (select-keys ds-data [:login :display_name])]
        (st/create-group! db (merge new-group group-data)))
      (throw (IllegalArgumentException. (format "No such group found via LDAP: %s" (:login new-group)))))))
